import subprocess
import functools
import contextlib
import traceback
import platform
import os
import logging
import pathlib

import psutil

from arps.core.real.rest_api_utils import (try_to_connect,
                                           build_url,
                                           wait_for_process,
                                           random_port)

from arps.apps.multiplatform_process import terminate_process


def remove_process_files(pid):
    logger = logging.getLogger()
    try:
        process = psutil.Process(pid)
    except psutil.NoSuchProcess:
        logger.warning('No such process %s', pid)
        return

    files = [pathlib.Path(pfile.path) for pfile in process.open_files()]
    files = [pfile for pfile in files if pfile.suffix == '.log']
    for f in files:
        try:
            print('Removing', f)
            os.remove(f)
        except PermissionError:
            logger.warning('Permission error when removing file %s: %s',
                           f, exec_info=True)
        except FileNotFoundError:
            logger.warning('File not found error when removing file %s: %s',
                           f, exec_info=True)
        except OSError as err:
            logger.warning(err, exec_info=True)


@contextlib.contextmanager
def start_agent_manager(user_defined_parameters):
    '''Start agent manager with user defined parameters, except port
    which is handled here

    It yields the resource endpoint partially built

    See agent_manager_runner argparse for more information about parameters.

    Args:
    - user_defined_parameters: parameters supported by agent_manager_runner
    - popen_parameters: subprocess parameters can be used to handle the agent manager process

    '''

    subprocess_popen = ['agent_manager_runner']
    subprocess_popen.extend(user_defined_parameters)

    port = random_port()
    subprocess_popen.extend(['--port', str(port)])

    with subprocess.Popen(subprocess_popen,
                          stderr=subprocess.PIPE,
                          stdout=subprocess.PIPE) as agent_manager:
        try:
            wait_for_process(agent_manager) #can raise RuntimeError or TimeoutError

            if try_to_connect(tries=5, port=port):
                yield functools.partial(build_url,
                                        f'{platform.node()}:{port}')

            remove_process_files(agent_manager.pid)
            process = psutil.Process(agent_manager.pid)
            for child in process.children():
                remove_process_files(child.pid)

            terminate_process(agent_manager)
            agent_manager.wait(timeout=5)
        except (RuntimeError, TimeoutError, subprocess.TimeoutExpired) as err:
            agent_manager.terminate()
            raise RuntimeError(err)
        except AssertionError:
            raise
        except Exception as err:
            print('Something unexpected happened', err)
            print('Exception type happened', type(err))
            print('traceback', traceback.format_exc())
        finally:
            if agent_manager.returncode is None:
                agent_manager.terminate()
            print('agent manager terminated {}'.format(agent_manager.returncode))


@contextlib.contextmanager
def start_pmt_service():
    pmt_port = random_port()
    pmt_popen = ['pmt_service','--port', str(pmt_port)]
    with subprocess.Popen(pmt_popen) as pmt:
        wait_for_process(pmt, timeout=30)
        assert try_to_connect(port=pmt_port)

        try:
            yield pmt_port

            remove_process_files(pmt.pid)
            terminate_process(pmt)
            pmt.wait(5)
        except (RuntimeError, TimeoutError, subprocess.TimeoutExpired) as err:
            raise RuntimeError(err)
        except AssertionError:
            raise
        except Exception as err:
            print('Something unexpected happened', err)
            print('Exception type happened', type(err))
            print('traceback', traceback.format_exc())
        finally:
            if pmt.returncode is None:
                pmt.terminate()
            print('pmt service terminated {}'.format(pmt.returncode))


@contextlib.contextmanager
def start_agents_directory():
    agents_directory_port = random_port()
    agents_directory_popen = ['agents_directory']

    agents_directory_popen.extend(['--port', str(agents_directory_port)])
    with subprocess.Popen(agents_directory_popen) as agents_directory:
        try:
            wait_for_process(agents_directory)
            if try_to_connect(port=agents_directory_port, tries=5):
                yield agents_directory_port

            remove_process_files(agents_directory.pid)
            terminate_process(agents_directory)
            agents_directory.wait(5)
        except (RuntimeError, TimeoutError, subprocess.TimeoutExpired) as err:
            print('Error while running agents directory', err)
            raise RuntimeError(err)
        except AssertionError:
            raise
        except Exception as err:
            print('Something unexpected happened', err)
            print('Exception type happened', type(err))
            print('traceback', traceback.format_exc())
        finally:
            if agents_directory.returncode is None:
                agents_directory.terminate()
            print('agents directory service terminated {}'.format(agents_directory.returncode))
